//=============================================================================
//Display Library
//for EA-DOGM132 GLCD (132px x 32px)
//and EA-DOGM128 GLCD (128px x 64px)
//
//Provides all basic functions to access the display
//Since no graphics ram is used, the memory footprint is rather small but also
//does not allow to change single pixels. Data written to the LCD can not be
//read back!
//Text can be displayed using the attached font generator containing several
//character sets and font variants.
//Thanks to Oliver Schwaneberg for adding several new functions to this library!
//
//Author:  Jan Michel (jan at mueschelsoft dot de)
//License: GNU General Public License, version 3
//Version: v0.91 January 2010
//=============================================================================
//New features in v0.91
// - Control of chip select pin
// - backlight & chip select are optional - can be turned of in header file
// - added function lcd_draw_image_P 
// - added function lcd_draw_image_xy_P
// - added function lcd_clear_area
// - added SPI initialization
//=============================================================================

#include "dogm-graphic.h"

//=============================================================================
//keeping track of current position in ram - necessary for big fonts & bitmaps
//=============================================================================

uint8_t lcd_current_page = 0;
uint8_t lcd_current_column = 0;


//change page address by s
uint8_t lcd_inc_page(int8_t s) {
  lcd_current_page += s;
  while (lcd_current_page >= LCD_RAM_PAGES) 
    lcd_current_page -= LCD_RAM_PAGES;
  return lcd_current_page;
  }

//change column address by s, 
uint8_t lcd_inc_column(int16_t s) {
  lcd_current_column += s;
#if LCD_WRAP_AROUND == 1
  if (lcd_current_column >= LCD_WIDTH) {
    if (s > 0) lcd_inc_page(1);
    else       lcd_inc_page(-1);
    while (lcd_current_column >= LCD_WIDTH) {
      if (s>0) lcd_current_column -= LCD_WIDTH;
      else     lcd_current_column += LCD_WIDTH;
      }
    }
#endif
  return lcd_current_column;
  }


  
  
//Move cursor to new position and update internal address
void lcd_moveto_xy(uint8_t page, uint8_t column) {
  LCD_GOTO_ADDRESS(page,column);
  lcd_current_column = column; 
  lcd_current_page = page;
  }

void lcd_move_xy(int8_t pages, int16_t columns) {
  lcd_moveto_xy(lcd_inc_page(pages),lcd_inc_column(columns));
  }


//=============================================================================
//Basic Byte Access to Display
//=============================================================================

//Writes one byte to the data ram
void lcd_data(uint8_t data) {
  LCD_SELECT();
  LCD_DRAM();
  spi_write(data);
  LCD_UNSELECT();
  lcd_inc_column(1);
  }

//Writes one command byte
void lcd_command(uint8_t cmd) {
  LCD_SELECT();
  LCD_CMD();
  spi_write(cmd);
  LCD_UNSELECT();
  }
  

//=============================================================================
//Puts raw data from Flash to the Display
//=============================================================================
#if LCD_INCLUDE_GRAPHIC_FUNCTIONS >= 1
/**
 * This function draws a bitmap from the current position on the screen.
 * Parameters:
 * progmem_image - prog_uint8_t array of columns aka the bitmap image
 * pages         - height of image in pages
 * columns       - width of image in pixels (or columns)
 * style         - Bit2: sets inverse mode
 */  
void lcd_draw_image_P(PGM_P progmem_image, uint8_t pages, uint8_t columns, uint8_t style) {
	uint8_t i,j;
  uint8_t inv = style & INVERT_BIT;
	for(j=0; j<pages && (lcd_get_position_page() < LCD_RAM_PAGES); j++) {
		for (i=0; i<columns && (lcd_get_position_column() < LCD_WIDTH); i++) {
      uint8_t tmp = pgm_read_byte(progmem_image++);
			if(!inv)
				lcd_data(tmp);
			else
				lcd_data(~tmp);
      }
    if(j+1 != pages)
      lcd_move_xy(1,-columns);
    }
  }

  
/**
 * This function draws a bitmap at any xy-position on the screen. 
 * Be aware that some pixels are deleted due to memory organization!
 * Parameters:
 * progmem_image - prog_uint8_t array of columns aka the bitmap image
 * x             - x start coordinate on the screen (in pixel)
 * y             - y start coordinate on the screen (in pixel)
 * pages         - height of image in pages
 * columns       - width of image in pixels
 * style         - Bit2: sets inverse mode
 */
void lcd_draw_image_xy_P(PGM_P progmem_image, uint8_t x, uint8_t y, uint8_t pages, uint8_t columns, uint8_t style) {
	uint16_t i,j;
  uint8_t data   = 0;
  uint8_t inv    = style & INVERT_BIT;
	uint8_t offset = y & 0x7; //Optimized modulo 8
	//If there is an offset, we must use an additional page
	if(offset)  
    pages++;
  //If there is not enough vertical space -> cut image
  if(pages > LCD_RAM_PAGES - lcd_get_position_page())   
    pages = LCD_RAM_PAGES - lcd_get_position_page();
  //Goto starting point and draw
	lcd_moveto_xy((y>>3), x);
	for (j=0; j<pages; j++) {
		for (i=0; i<columns && (lcd_get_position_column() < LCD_WIDTH); i++){
      data = 0;
      if (!offset || j+1 != pages)
        data = pgm_read_byte(&progmem_image[j*columns + i]) << offset;
			if(j > 0 && offset)
				data |= pgm_read_byte(&progmem_image[(j-1)*columns + i]) >> (8-offset);
			if(inv)	lcd_data(~data);
      else 		lcd_data(data);
		  }
		if(j+1 != pages)
			lcd_move_xy(1,-columns);
	  }
  }
#endif


/**
 * This function clears an area of the screen
 * pages         - height of area in pages
 * columns       - width of area in pixels
 * style         - Bit2: sets inverse mode
 * Cursor is moved to start of area after clear
 */
void lcd_clear_area(uint8_t pages, uint8_t columns, uint8_t style) {
  uint8_t i,j,max;
  uint8_t inv = (style & INVERT_BIT)?0xFF:0;
  
  if(pages > (max = LCD_RAM_PAGES - lcd_get_position_page()))   
    pages = max;
  if(columns > (max = LCD_WIDTH - lcd_get_position_column()))   
    columns = max;
  
  for(j=0; j<pages; j++) {
    for(i=0; i<columns; i++) {
      lcd_data(inv);
      }
    lcd_move_xy(1,-columns);
    }
  lcd_move_xy(-pages,0);
  }


/**
  * Initializes the display in 6 o'clock mode, 4x booster for 2.4-3.3V supply voltage
  * scheme according to datasheet
  * Compatible to both Display types: 128x64 and 132x32
  */
void lcd_init() {

  //set outputs
  LCD_SET_PIN_DIRECTIONS();
  //Initialize SPI Interface
  LCD_INIT_SPI();   
  //Apply Reset to the Display Controller
  LCD_RESET();
  //Load settings
  LCD_SET_BOTTOM_VIEW();              //6 o'clock mode, normal orientation
  LCD_ORIENTATION_NORMAL();
  LCD_SET_MODE_POSITIVE();            //positive display
  LCD_SET_BIAS_RATIO_1_9();           //bias 1/9
  LCD_SET_POWER_CONTROL(7);           //power control mode: all features on
  LCD_SET_BIAS_VOLTAGE(3);            //set voltage regulator R/R
  LCD_SET_VOLUME_MODE(0x1F);          //volume mode set
  LCD_SET_INDICATOR_OFF();            //switch indicator off, no blinking
  LCD_SWITCH_ON();                    //Switch display on
  uint8_t i,j;                        //clear display content
  for(j=0;j<LCD_RAM_PAGES;j++) {
    lcd_moveto_xy(j,0);
    for (i=0;i<LCD_WIDTH;i++)
      lcd_data(0);
    }
  lcd_moveto_xy(0,0);                 //Set write pointer
  LCD_SET_FIRST_LINE(0);              //first bit in RAM is on the first line of the LCD
  return;
  }
